from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

from fretboardgtr.elements.base import FretBoardElement
from fretboardgtr.elements.notes import FrettedNote, OpenNote
from fretboardgtr.fretboards.config import FretBoardConfig
from fretboardgtr.fretboards.elements import FretBoardElements
from fretboardgtr.notes_creators import NotesContainer


class FretBoardLike(ABC):
    """Interface to implement to define a FretBoard.

    This interface's purpose is to define an interface (common
    behaviour) in order to draw the fretboard into a SVG Format. This
    interface also define standard method for higher level API call such
    as add_note_element, or add_element for example.
    """

    @abstractmethod
    def set_config(self, config: FretBoardConfig) -> None:
        pass

    @abstractmethod
    def add_note_element(self, note: Union[OpenNote, FrettedNote]) -> None:
        pass

    @abstractmethod
    def add_element(self, element: FretBoardElement) -> None:
        pass

    @abstractmethod
    def add_fingering(
        self, fingering: List[Optional[int]], root: Optional[str] = None
    ) -> None:
        pass

    @abstractmethod
    def add_notes(self, scale: NotesContainer) -> None:
        pass

    @abstractmethod
    def get_size(self) -> Tuple[float, float]:
        pass

    @abstractmethod
    def get_inside_bounds(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        pass

    @abstractmethod
    def get_elements(self) -> FretBoardElements:
        pass



import logging
from typing import Any, Dict


class ConfigIniter:
    """Mixin class that define methods read the different configurations.

    All the components/elements configuration must subclass this Mixin.
    Then, all the configuration can be read thank to the method (eg
    from_dict) recursively
    """

    @classmethod
    def from_dict(cls, _dict: Dict) -> Any:
        kwargs = {}
        for arg in _dict:
            if arg not in cls.__annotations__.keys():
                logging.warning(
                    f'"{arg}" key confuration was not found in {cls.__name__}'
                )
                continue
            _type = cls.__annotations__[arg]
            if hasattr(_type, "from_dict"):
                kwargs[arg] = _type.from_dict(_dict[arg])
            else:
                kwargs[arg] = _dict[arg]
        return cls(**kwargs)



from dataclasses import dataclass

from fretboardgtr.constants import INTERVAL_MAPPING, WHITE

TEXT_OFFSET = "0.3em"
TEXT_STYLE = "text-anchor:middle"
from fretboardgtr.base import ConfigIniter


@dataclass
class NoteColors(ConfigIniter):
    """Dataclass containing the mapping of colors and intervals."""

    root: str = "rgb(231, 0, 0)"
    minor_second: str = "rgb(249, 229, 0)"
    major_second: str = "rgb(249, 165, 0)"
    minor_third: str = "rgb(0, 94, 0)"
    major_third: str = "rgb(0, 108, 0)"
    perfect_fourth: str = "rgb(0, 154, 0)"
    diminished_fifth: str = "rgb(0, 15, 65)"
    perfect_fifth: str = "rgb(0, 73, 151)"
    minor_sixth: str = "rgb(168, 107, 98)"
    major_sixth: str = "rgb(222, 81, 108)"
    minor_seventh: str = "rgb(120, 37, 134)"
    major_seventh: str = "rgb(120, 25, 98)"

    def from_short_interval(self, interval: str) -> str:
        """Get color for the given short interval.

        Parameters
        ----------
        interval : str
            String representing the interval

        Returns
        -------
        str
            RGB color as string

        Example
        -------
        from fretboardgtr.constants import Interval
        >>> NoteColors().from_short_interval(Interval.MINOR_SIXTH)
        "rgb(168, 107, 98)"
        """
        color = WHITE
        for long, short in INTERVAL_MAPPING.items():
            if interval != short:
                continue
            if hasattr(self, long):
                color = getattr(self, long)
        return color

    def from_interval(self, interval: int) -> str:
        """Get color for the given long interval name.

        Parameters
        ----------
        interval : str
            String representing the long interval

        Returns
        -------
        str
            RGB color as string

        Example
        -------
        from fretboardgtr.constants import LongInterval
        >>> NoteColors().from_short_interval(LongInterval.MINOR_SIXTH)
        "rgb(168, 107, 98)"
        """
        cls_keys = list(self.__annotations__)
        color = getattr(self, cls_keys[interval % 12])
        return color



from typing import Any, List, Optional, Tuple

from fretboardgtr.constants import STANDARD_TUNING
from fretboardgtr.fretboards.config import FretBoardConfig


class VerticalFretBoard:
    """Class containing the different elements of a fretboard.

    Also contains associated method to add some.
    """

    def __init__(
        self,
        tuning: Optional[List[str]] = None,
        config: Optional[FretBoardConfig] = None,
    ):
        self.tuning = tuning if tuning is not None else STANDARD_TUNING
        self.config = config if config is not None else FretBoardConfig()

    def get_list_in_good_order(self, _list: List[Any]) -> List[Any]:
        return _list

    def get_background_start_position(self) -> Tuple[float, float]:
        open_fret_height = self.config.general.fret_height
        return (
            self.config.general.x_start,
            self.config.general.y_start + open_fret_height,
        )

    def get_background_dimensions(self) -> Tuple[float, float]:
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret
        width = (len(self.tuning) - 1) * self.config.general.fret_width
        height = (number_of_frets) * (self.config.general.fret_height)
        return width, height

    def get_neck_dot_position(self, dot: int) -> List[Tuple[float, float]]:
        x = (
            self.config.general.x_start
            + (len(self.tuning) / 2 - (1 / 2)) * self.config.general.fret_width
        )
        y = (
            self.config.general.y_start
            + (0.5 + dot - self.config.general.first_fret)
            * self.config.general.fret_height
        )
        if dot % 12 == 0:
            # Add two dots dot is multiple of 12
            lower_position = (
                x - self.config.general.fret_width,
                y,
            )
            upper_position = (
                x + self.config.general.fret_width,
                y,
            )
            return [lower_position, upper_position]
        else:
            center_position = (
                x,
                y,
            )
        return [center_position]

    def get_fret_position(
        self, fret_no: int
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        y = self.config.general.y_start + (self.config.general.fret_height) * (fret_no)
        y_start = self.config.general.x_start
        x_end = self.config.general.x_start + (self.config.general.fret_width) * (
            len(self.tuning) - 1
        )
        return (y_start, y), (x_end, y)

    def get_strings_position(
        self, string_no: int
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        open_fret_height = self.config.general.fret_height

        y_start = self.config.general.y_start + open_fret_height
        y_end = self.config.general.y_start + (
            self.config.general.fret_height
            + (self.config.general.last_fret - self.config.general.first_fret)
            * self.config.general.fret_height
        )
        start = (
            self.config.general.x_start
            + (self.config.general.fret_width) * (string_no),
            y_start,
        )
        end = (
            self.config.general.x_start
            + (self.config.general.fret_width) * (string_no),
            y_end,
        )
        return start, end

    def get_nut_position(
        self,
    ) -> Optional[Tuple[Tuple[float, float], Tuple[float, float]]]:
        if self.config.general.first_fret != 0 or not self.config.general.show_nut:
            return None
        open_fret_height = self.config.general.fret_height

        start = (
            self.config.general.x_start,
            self.config.general.y_start + open_fret_height,
        )
        end = (
            self.config.general.x_start
            + self.config.general.fret_width * (len(self.tuning) - 1),
            self.config.general.y_start + open_fret_height,
        )
        return start, end

    def get_fret_number_position(self, dot: int) -> Tuple[float, float]:
        x = self.config.general.x_start + self.config.general.fret_width * (
            len(self.tuning)
        )
        y = self.config.general.y_start + self.config.general.fret_height * (
            1 / 2 + dot - self.config.general.first_fret
        )
        return x, y

    def get_tuning_position(self, string_no: int) -> Tuple[float, float]:
        x = self.config.general.x_start + self.config.general.fret_width * (string_no)
        y = self.config.general.y_start + (
            self.config.general.fret_height
            * (self.config.general.last_fret - self.config.general.first_fret + 3 / 2)
        )
        return x, y

    def get_single_note_position(
        self, string_no: int, index: int
    ) -> Tuple[float, float]:
        y_pos = index + (1 / 2)
        x = self.config.general.x_start + self.config.general.fret_width * (string_no)
        y = self.config.general.y_start + (self.config.general.fret_height) * y_pos
        return x, y

    def get_cross_position(self, string_no: int) -> Tuple[float, float]:
        x = self.config.general.x_start + (self.config.general.fret_width) * (string_no)
        y = self.config.general.y_start + self.config.general.fret_height * (1 / 2)
        return x, y

    def get_size(self) -> Tuple[float, float]:
        """Get total size of the drawing.

        Returns
        -------
        Tuple[float, float]
            Width and heigth
        """
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret
        width = (
            self.config.general.x_start
            + self.config.general.fret_width * (len(self.tuning) + 1)
            + self.config.general.y_end_offset
        )
        height = (
            self.config.general.y_start
            + self.config.general.fret_height * (number_of_frets + 2)
            + self.config.general.x_end_offset
        )
        return (width, height)

    def get_inside_bounds(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Get the size of the inner drawing.

        This function could be use to add custom elements

        Returns
        -------
        Tuple[Tuple[float, float], Tuple[float, float]]
            Upper left corner x coordinate, upper left corner y coordinate
            Lower right corner x coordinate, lower right corner y coordinate
        """
        open_fret_height = self.config.general.fret_height
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret
        upper_left_x = self.config.general.x_start
        upper_left_y = self.config.general.y_start
        lower_right_x = (
            self.config.general.x_start
            + (len(self.tuning) - 1) * self.config.general.fret_width
        )
        lower_right_y = (
            self.config.general.y_start
            + open_fret_height
            + (number_of_frets) * (self.config.general.fret_height)
        )
        return ((upper_left_x, upper_left_y), (lower_right_x, lower_right_y))



import mimetypes
import tempfile
from pathlib import Path

import pytest
import svgwrite
from PIL import Image
from pypdf import PdfReader

from fretboardgtr.exporters import PDFExporter, PNGExporter, SVGExporter


@pytest.fixture()
def drawing():
    dwg = svgwrite.Drawing(
        size=(  # +2 == Last fret + tuning
            100,
            100,
        ),
        profile="full",
    )
    circle = svgwrite.shapes.Circle(
        (50, 50),
        r=20,
    )
    dwg.add(circle)
    return dwg


def test_svg_exporter(drawing):
    svg_exporter = SVGExporter(drawing)
    with tempfile.TemporaryDirectory() as tmp_dir:
        outfile = Path(tmp_dir) / "tmp_file.svg"
        assert not outfile.exists()
        svg_exporter.export(outfile)
        assert outfile.exists()
        assert mimetypes.guess_type(outfile)[0] == "image/svg+xml"


def test_png_exporter(drawing):
    png_exporter = PNGExporter(drawing)
    with tempfile.TemporaryDirectory() as tmp_dir:
        outfile = Path(tmp_dir) / "tmp_file.png"
        assert not outfile.exists()
        png_exporter.export(outfile)
        assert outfile.exists()
        assert mimetypes.guess_type(outfile)[0] == "image/png"
        with Image.open(outfile) as img:
            assert img.format == "PNG"


def test_pdf_exporter(drawing):
    pdf_exporter = PDFExporter(drawing)
    with tempfile.TemporaryDirectory() as tmp_dir:
        outfile = Path(tmp_dir) / "tmp_file.pdf"
        assert not outfile.exists()
        pdf_exporter.export(outfile)
        assert outfile.exists()
        assert mimetypes.guess_type(outfile)[0] == "application/pdf"
        assert len(PdfReader(outfile).pages) == 1



import tempfile
import uuid
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Type, Union

import svgwrite


class Exporter(ABC):
    """Interface to implement for a new exporter."""

    def __init__(self, drawing: svgwrite.Drawing):
        self.drawing = drawing

    @abstractmethod
    def export(self, to: Union[str, Path]) -> None:
        pass


class SVGExporter(Exporter):
    """SVG Exporter."""

    def export(self, to: Union[str, Path]) -> None:
        to = Path(to)
        self.drawing.saveas(str(to))


class PNGExporter(Exporter):
    """PNG Exporter.

    Need reportlab and svglib module installed
    """

    def export(self, to: Union[str, Path]) -> None:
        try:
            from reportlab.graphics import renderPM
        except ImportError:
            raise ImportError(
                "Cannot export svg to PNG because reportlab package is missing"
            )
        try:
            from svglib.svglib import svg2rlg
        except ImportError:
            raise ImportError(
                "Cannot export svg to PNG because svglib package is missing"
            )
        to = Path(to)
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_file = Path(tmp_dir) / Path(f"{uuid.uuid4()}.svg")
            self.drawing.saveas(str(tmp_file))
            drawing = svg2rlg(str(tmp_file))
            renderPM.drawToFile(drawing, str(to), fmt="PNG")


class PDFExporter(Exporter):
    """PDF Exporter.

    Need reportlab and svglib module installed
    """

    def export(self, to: Union[str, Path]) -> None:
        try:
            from reportlab.graphics import renderPDF
        except ImportError:
            raise ImportError(
                "Cannot export svg to PNG because reportlab package is missing"
            )
        try:
            from svglib.svglib import svg2rlg
        except ImportError:
            raise ImportError(
                "Cannot export svg to PNG because svglib package is missing"
            )
        to = Path(to)
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_file = Path(tmp_dir) / Path(f"{uuid.uuid4()}.svg")
            self.drawing.saveas(str(tmp_file))

            drawing = svg2rlg(str(tmp_file))
            renderPDF.drawToFile(drawing, str(to))


def register_exporter(exporter: Type[Exporter], extension: str) -> None:
    """Register an exporter.

    When creating a new exporter one have to use this function to
    register the exporter and be able to use it
    """
    EXPORTERS[extension.upper()] = exporter


EXPORTERS: Dict[str, Type[Exporter]] = {}

register_exporter(SVGExporter, "SVG")
register_exporter(PNGExporter, "PNG")
register_exporter(PDFExporter, "PDF")



from fretboardgtr.elements.neck_dots import NeckDot, NeckDotConfig


def test_neck_dots_get_svg():
    neck_dot = NeckDot(position=(0.0, 0.0))
    attribs = neck_dot.get_svg().attribs
    assert attribs["cx"] == 0.0
    assert attribs["cy"] == 0.0


def test_neck_dots_get_svg_custom_config():
    neck_dot_config = NeckDotConfig(radius=30)
    neck_dot = NeckDot(position=(0.0, 0.0), config=neck_dot_config)
    attribs = neck_dot.get_svg().attribs
    assert attribs["cx"] == 0.0
    assert attribs["cy"] == 0.0
    assert attribs["r"] == 30



import pytest

from fretboardgtr.elements.background import Background, BackgroundConfig
from fretboardgtr.elements.fret_number import FretNumber, FretNumberConfig
from fretboardgtr.elements.frets import Fret, FretConfig
from fretboardgtr.elements.neck_dots import NeckDot, NeckDotConfig
from fretboardgtr.elements.notes import (
    FrettedNote,
    FrettedNoteConfig,
    OpenNote,
    OpenNoteConfig,
)
from fretboardgtr.elements.nut import Nut, NutConfig
from fretboardgtr.elements.strings import String, StringConfig
from fretboardgtr.elements.tuning import Tuning, TuningConfig
from fretboardgtr.fretboard import FretBoard
from fretboardgtr.fretboards.config import FretBoardConfig, FretBoardGeneralConfig
from fretboardgtr.note_colors import NoteColors
from fretboardgtr.notes_creators import NotesContainer


@pytest.fixture()
def default_config():
    return FretBoardConfig(
        general=FretBoardGeneralConfig(
            x_start=30.0,
            y_start=30.0,
            fret_height=50,
            fret_width=70,
            first_fret=0,
            last_fret=12,
            show_tuning=True,
            show_frets=True,
            show_nut=True,
            show_degree_name=False,
            show_note_name=True,
            open_color_scale=False,
            fretted_color_scale=True,
            open_colors=NoteColors(
                root="rgb(231, 0, 0)",
                minor_second="rgb(249, 229, 0)",
                major_second="rgb(249, 165, 0)",
                minor_third="rgb(0, 94, 0)",
                major_third="rgb(0, 108, 0)",
                perfect_fourth="rgb(0, 154, 0)",
                diminished_fifth="rgb(0, 15, 65)",
                perfect_fifth="rgb(0, 73, 151)",
                minor_sixth="rgb(168, 107, 98)",
                major_sixth="rgb(222, 81, 108)",
                minor_seventh="rgb(120, 37, 134)",
                major_seventh="rgb(120, 25, 98)",
            ),
            fretted_colors=NoteColors(
                root="rgb(231, 0, 0)",
                minor_second="rgb(249, 229, 0)",
                major_second="rgb(249, 165, 0)",
                minor_third="rgb(0, 94, 0)",
                major_third="rgb(0, 108, 0)",
                perfect_fourth="rgb(0, 154, 0)",
                diminished_fifth="rgb(0, 15, 65)",
                perfect_fifth="rgb(0, 73, 151)",
                minor_sixth="rgb(168, 107, 98)",
                major_sixth="rgb(222, 81, 108)",
                minor_seventh="rgb(120, 37, 134)",
                major_seventh="rgb(120, 25, 98)",
            ),
            enharmonic=True,
        ),
        background=BackgroundConfig(color="none", opacity=0.7),
        fret_numbers=FretNumberConfig(
            color="rgb(150,150,150)", fontsize=20, fontweight="bold"
        ),
        neck_dots=NeckDotConfig(
            color="rgb(200,200,200)",
            color_stroke="rgb(0,0,0)",
            width_stroke=2,
            radius=7,
        ),
        frets=FretConfig(color="rgb(150,150,150)", width=3),
        nut=NutConfig(color="rgb(0,0,0)", width=6),
        tuning=TuningConfig(
            color="rgb(150,150,150)",
            fontsize=20,
            fontweight="normal",
        ),
        strings=StringConfig(color="rgb(0,0,0)", width=3),
        open_notes=OpenNoteConfig(
            radius=20,
            color="rgb(255,255,255)",
            stroke_color="rgb(0,0,0)",
            stroke_width=3,
            text_color="rgb(0,0,0)",
            fontsize=20,
            fontweight="bold",
        ),
        fretted_notes=FrettedNoteConfig(
            radius=20,
            color="rgb(255,255,255)",
            stroke_color="rgb(0,0,0)",
            stroke_width=3,
            text_color="rgb(0,0,0)",
            fontsize=20,
            fontweight="bold",
        ),
    )


@pytest.fixture()
def dict_config():
    return dict(
        general=dict(
            x_start=30.0,
            y_start=30.0,
            fret_height=50,
            fret_width=70,
            first_fret=0,
            last_fret=12,
            show_tuning=True,
            show_frets=True,
            show_nut=True,
            show_degree_name=False,
            show_note_name=True,
            open_color_scale=False,
            fretted_color_scale=True,
            open_colors=dict(
                root="rgb(231, 0, 0)",
                minor_second="rgb(249, 229, 0)",
                major_second="rgb(249, 165, 0)",
                minor_third="rgb(0, 94, 0)",
                major_third="rgb(0, 108, 0)",
                perfect_fourth="rgb(0, 154, 0)",
                diminished_fifth="rgb(0, 15, 65)",
                perfect_fifth="rgb(0, 73, 151)",
                minor_sixth="rgb(168, 107, 98)",
                major_sixth="rgb(222, 81, 108)",
                minor_seventh="rgb(120, 37, 134)",
                major_seventh="rgb(120, 25, 98)",
            ),
            fretted_colors=dict(
                root="rgb(231, 0, 0)",
                minor_second="rgb(249, 229, 0)",
                major_second="rgb(249, 165, 0)",
                minor_third="rgb(0, 94, 0)",
                major_third="rgb(0, 108, 0)",
                perfect_fourth="rgb(0, 154, 0)",
                diminished_fifth="rgb(0, 15, 65)",
                perfect_fifth="rgb(0, 73, 151)",
                minor_sixth="rgb(168, 107, 98)",
                major_sixth="rgb(222, 81, 108)",
                minor_seventh="rgb(120, 37, 134)",
                major_seventh="rgb(120, 25, 98)",
            ),
            enharmonic=True,
        ),
        background=dict(color="rgb(150,150,150)", opacity=0.2),
        fret_numbers=dict(color="rgb(150,150,150)", fontsize=20, fontweight="bold"),
        neck_dots=dict(
            color="rgb(200,200,200)",
            color_stroke="rgb(0,0,0)",
            width_stroke=2,
            radius=7,
        ),
        frets=dict(color="rgb(150,150,150)", width=3),
        nut=dict(color="rgb(0,0,0)", width=6),
        tuning=dict(
            color="rgb(150,150,150)",
            fontsize=20,
            fontweight="normal",
        ),
        strings=dict(color="rgb(0,0,0)", width=3),
        open_notes=dict(
            radius=20,
            color="rgb(255,255,255)",
            stroke_color="rgb(0,0,0)",
            stroke_width=3,
            text_color="rgb(0,0,0)",
            fontsize=20,
            fontweight="bold",
        ),
        fretted_notes=dict(
            radius=20,
            color="rgb(255,255,255)",
            stroke_color="rgb(0,0,0)",
            stroke_width=3,
            text_color="rgb(0,0,0)",
            fontsize=20,
            fontweight="bold",
        ),
    )


@pytest.fixture()
def partial_dict_config():
    return dict(
        background=dict(color="blue"),
    )


def test_default_config(default_config: FretBoardConfig):
    assert default_config == FretBoardConfig()


def load_config_from_dict(dict_config):
    config = FretBoardConfig.from_dict(dict_config)
    assert config == FretBoardConfig()


def load_config_from_partial_dict(partial_dict_config):
    config = FretBoardConfig.from_dict(partial_dict_config)
    assert config.background.color == "blue"
    # The default is still there
    assert config.background.opacity == 0.2


def test_fretboard_get_background(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_background()
    background = fretboard.elements.background
    # Position == (width_fret + x_start, y_start)
    assert background.position == (100, 30)
    assert background.size == (840, 250)


def test_fretboard_get_background_new_tuning(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config, tuning=["E", "A", "D", "G"])
    fretboard.add_background()
    background = fretboard.elements.background
    # Position == (width_fret + x_start, y_start)
    assert background.position == (100, 30)
    assert background.size == (840, 150)


def test_fretboard_get_neck_dots(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_neck_dots()
    neck_dots = fretboard.elements.neck_dots
    assert neck_dots[0].x == 275.0
    assert neck_dots[0].y == 155.0
    for neck_dot in neck_dots:
        # Check if center of fret
        assert (
            neck_dot.x
            % (
                fretboard.config.general.x_start
                - (fretboard.config.general.fret_width // 2)
            )
            == 0
        )
        assert (
            neck_dot.y
            % (
                fretboard.config.general.y_start
                - (fretboard.config.general.fret_height // 2)
            )
            == 0
        )


def test_fretboard_get_frets(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_frets()
    frets = fretboard.elements.frets
    for fret in frets:
        # Test if multiple of width
        assert (
            fret.start_position[0] % fretboard.config.general.x_start
            + fretboard.config.general.fret_width
        )
        # Test if vertical line
        assert fret.start_position[0] == fret.end_position[0]

        # Test coherence and length
        assert fret.start_position[1] == fretboard.config.general.y_start
        assert fret.end_position[1] == (
            fretboard.config.general.y_start
            + (len(fretboard.tuning) - 1) * fretboard.config.general.fret_height
        )


def test_fretboard_get_strings(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_strings()
    strings = fretboard.elements.strings
    for string in strings:
        # Test if horizontal line
        assert string.start_position[1] == string.end_position[1]

        # Test if multiple of width
        assert (
            string.start_position[1] % fretboard.config.general.x_start
            + fretboard.config.general.fret_height
        )

        # Test coherence and length
        assert (
            string.start_position[0]
            == fretboard.config.general.x_start + fretboard.config.general.fret_width
        )
        assert string.end_position[0] == (
            fretboard.config.general.y_start
            + fretboard.config.general.fret_width
            + fretboard.config.general.last_fret * fretboard.config.general.fret_width
        )


def test_fretboard_get_nut(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_nut()
    nut = fretboard.elements.nut
    assert nut is not None
    # Test if vertical line
    assert nut.start_position[0] == nut.end_position[0]
    # Test if start where it should
    assert (
        nut.start_position[0]
        == fretboard.config.general.x_start + fretboard.config.general.fret_width
    )
    # Test if start where it should
    assert nut.start_position[1] == fretboard.config.general.y_start
    # Test length
    assert (
        nut.end_position[1]
        == fretboard.config.general.y_start
        + (len(fretboard.tuning) - 1) * fretboard.config.general.fret_height
    )


def test_fretboard_get_nut_not_first_fret(default_config: FretBoardConfig):
    default_config.general.first_fret = 1
    fretboard = FretBoard(config=default_config)
    fretboard.add_nut()

    assert fretboard.elements.nut is None


def test_fretboard_get_fret_numbers(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_fret_numbers()
    fret_numbers = fretboard.elements.fret_numbers
    assert fret_numbers is not None
    assert fret_numbers[0].x == 275.0
    assert fret_numbers[0].y == 330.0
    for fret_number in fret_numbers:
        # Check if center of fret
        assert (
            fret_number.x
            % (
                fretboard.config.general.x_start
                - (fretboard.config.general.fret_width // 2)
            )
            == 0
        )
        assert (
            fret_number.y
            % (
                fretboard.config.general.y_start
                - (fretboard.config.general.fret_height // 2)
            )
            == 0
        )


def test_fretboard_get_string(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    strings = fretboard.add_strings()
    strings = fretboard.elements.strings
    for string in strings:
        # Test if horizontal line
        assert string.start_position[1] == string.end_position[1]

        # Test if multiple of width
        assert (
            string.start_position[1] % fretboard.config.general.x_start
            + fretboard.config.general.fret_height
        )

        # Test coherence and length
        assert (
            string.start_position[0]
            == fretboard.config.general.x_start + fretboard.config.general.fret_width
        )
        assert string.end_position[0] == (
            fretboard.config.general.y_start
            + fretboard.config.general.fret_width
            + fretboard.config.general.last_fret * fretboard.config.general.fret_width
        )


def test_fretboard_get_open_note(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_open_note(position=(100, 100), note="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "C"
    assert open_note.config.color == "rgb(255,255,255)"


def test_fretboard_get_open_note_root(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_open_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "C"
    assert open_note.config.color == "rgb(255,255,255)"


def test_fretboard_get_open_note_root_colors(default_config: FretBoardConfig):
    default_config.general.open_color_scale = True
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_open_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "C"
    # RED
    assert open_note.config.color == "rgb(231, 0, 0)"


def test_fretboard_get_open_note_show_degree_name(default_config: FretBoardConfig):
    default_config.general.open_color_scale = True
    default_config.general.show_degree_name = True
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_open_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "1"
    # RED
    assert open_note.config.color == "rgb(231, 0, 0)"


def test_fretboard_get_open_note_no_name(default_config: FretBoardConfig):
    default_config.general.open_color_scale = True
    default_config.general.show_degree_name = False
    default_config.general.show_note_name = False
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_open_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == ""
    # RED
    assert open_note.config.color == "rgb(231, 0, 0)"


def test_fretboard_get_fretted_note(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_fretted_note(position=(100, 100), note="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "C"
    assert open_note.config.color == "rgb(255,255,255)"


def test_fretboard_get_fretted_note_root_colors(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_fretted_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "C"
    # RED
    assert open_note.config.color == "rgb(231, 0, 0)"


def test_fretboard_get_fretted_note_root_no_color(default_config: FretBoardConfig):
    default_config.general.fretted_color_scale = False
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_fretted_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "C"
    assert open_note.config.color == "rgb(255,255,255)"


def test_fretboard_get_fretted_note_show_degree_name(default_config: FretBoardConfig):
    default_config.general.show_degree_name = True
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_fretted_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == "1"
    # RED
    assert open_note.config.color == "rgb(231, 0, 0)"


def test_fretboard_get_fretted_note_no_name(default_config: FretBoardConfig):
    default_config.general.show_degree_name = False
    default_config.general.show_note_name = False
    fretboard = FretBoard(config=default_config)
    open_note = fretboard._get_fretted_note(position=(100, 100), note="C", root="C")
    assert open_note.x == 100
    assert open_note.y == 100
    assert open_note.name == ""
    # RED
    assert open_note.config.color == "rgb(231, 0, 0)"


def test_fretboard_get_notes(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    fretboard.add_note(string_no=0, note="E", root="C")

    notes = fretboard.elements.notes
    for note in notes:
        assert note.y == fretboard.config.general.y_start
        # Check if placed between two frets
        assert (
            note.x
            % (
                fretboard.config.general.x_start
                - (fretboard.config.general.fret_width // 2)
            )
            == 0
        )


def test_fretboard_get_notes_invalid_string(default_config: FretBoardConfig):
    fretboard = FretBoard(config=default_config)
    with pytest.raises(ValueError):
        fretboard.add_note(string_no=7, note="E", root="C")


def test_fretboard_get_inside_bounds(default_config):
    fretboard = FretBoard(config=default_config)
    upper_left, lower_right = fretboard.get_inside_bounds()
    assert upper_left == (30.0, 30.0)
    assert lower_right == (940.0, 280.0)


def test_fretboard_get_size(default_config):
    fretboard = FretBoard(config=default_config)
    width, height = fretboard.get_size()
    assert width == 1010.0
    assert height == 380.0


def test_init_fretboard(default_config):
    fretboard = FretBoard(config=default_config)
    assert len(fretboard.elements) == 38
    assert len(fretboard.elements.strings) == 6
    assert len(fretboard.elements.frets) == 13
    list_of_elements = fretboard.elements.to_list()
    assert any([isinstance(obj, FretNumber) for obj in list_of_elements])
    assert any([isinstance(obj, Fret) for obj in list_of_elements])
    assert any([isinstance(obj, NeckDot) for obj in list_of_elements])
    assert any([isinstance(obj, Nut) for obj in list_of_elements])
    assert any([isinstance(obj, String) for obj in list_of_elements])
    assert any([isinstance(obj, Tuning) for obj in list_of_elements])
    assert not any([isinstance(obj, FrettedNote) for obj in list_of_elements])
    assert not any([isinstance(obj, OpenNote) for obj in list_of_elements])


def test_init_fretboard_add_notes(default_config):
    fretboard = FretBoard(config=default_config)
    notes_container = NotesContainer(
        root="C", notes=["C", "D", "E", "F", "G", "A", "B"]
    )
    fretboard.add_notes(notes_container)
    assert len(fretboard.elements) == 86
    assert len(fretboard.elements.strings) == 6
    assert len(fretboard.elements.frets) == 13
    list_of_elements = fretboard.elements.to_list()
    assert any([isinstance(obj, Background) for obj in list_of_elements])
    assert any([isinstance(obj, FretNumber) for obj in list_of_elements])
    assert any([isinstance(obj, Fret) for obj in list_of_elements])
    assert any([isinstance(obj, NeckDot) for obj in list_of_elements])
    assert any([isinstance(obj, Nut) for obj in list_of_elements])
    assert any([isinstance(obj, String) for obj in list_of_elements])
    assert any([isinstance(obj, Tuning) for obj in list_of_elements])
    assert any([isinstance(obj, FrettedNote) for obj in list_of_elements])
    assert any([isinstance(obj, OpenNote) for obj in list_of_elements])


def test_add_element(default_config):
    note = OpenNote("C", position=(0, 0))
    fretboard = FretBoard(config=default_config)
    fretboard.add_element(note)


def test_add_wrong_lement(default_config):
    fretboard = FretBoard(config=default_config)
    with pytest.raises(ValueError):
        fretboard.add_element(1)



from fretboardgtr.elements.nut import Nut, NutConfig


def test_nut_get_svg():
    nut = Nut(start_position=(0.0, 0.0), end_position=(10.0, 10.0))
    attribs = nut.get_svg().attribs
    print(attribs)
    assert attribs["x1"] == 0.0
    assert attribs["y1"] == 0.0
    assert attribs["x2"] == 10.0
    assert attribs["y2"] == 10.0


def test_nut_get_svg_custom_config():
    nut_config = NutConfig(color="blue")
    nut = Nut(start_position=(0.0, 0.0), end_position=(10.0, 10.0), config=nut_config)
    attribs = nut.get_svg().attribs
    assert attribs["x1"] == 0.0
    assert attribs["y1"] == 0.0
    assert attribs["x2"] == 10.0
    assert attribs["y2"] == 10.0
    assert attribs["stroke"] == "blue"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

from docs.scripts.plot import make_plots

BASE_URL = "https://github.com/antscloud/fretboardgtr"
# -- Project information -----------------------------------------------------

project = "fretboardgtr"
copyright = "Antoine Gibek"
author = "Antscloud"

LOGO_PATH = os.path.join(os.path.dirname(__file__), "assets/black_logo_square@2x.png")
# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
    "myst_parser",
]


def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return f"{BASE_URL}/tree/main/{filename}.py"


autosummary_generate = True

napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True
myst_heading_anchors = 3

autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

# -- Options for autodoc ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

html_sidebars = {
    "**": [
        "navbar-logo.html",
        "search-field.html",
        "sbt-sidebar-nav.html",
    ]
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"

html_theme_options = {
    "repository_url": BASE_URL,
    "use_repository_button": True,
    "use_issues_button": True,
    "use_download_button": True,
    "use_sidenotes": True,
    "home_page_in_toc": False,
    "show_toc_level": 2,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["assets", "_static"]

html_css_files = [
    "css/custom.css",
]

html_logo = LOGO_PATH
html_favicon = LOGO_PATH
html_title = "FretBoardGtr"

# -- Custom code to run -------------------------------------------------

make_plots()



from dataclasses import dataclass
from typing import Optional, Tuple

import svgwrite

from fretboardgtr.base import ConfigIniter
from fretboardgtr.constants import BLACK
from fretboardgtr.elements.base import FretBoardElement


@dataclass
class StringConfig(ConfigIniter):
    """String element configuration."""

    color: str = BLACK
    width: int = 3


class String(FretBoardElement):
    """String elements to be drawn in the final fretboard."""

    def __init__(
        self,
        start_position: Tuple[float, float],
        end_position: Tuple[float, float],
        width: Optional[int] = None,
        config: Optional[StringConfig] = None,
    ):
        self.config = config if config else StringConfig()
        self.start_position = start_position
        self.end_position = end_position
        self.width = width

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the String to a svgwrite object.

        This maps the StringConfig configuration attributes to the svg
        attributes
        """
        if self.width is None:
            self.width = self.config.width

        line = svgwrite.shapes.Line(
            start=self.start_position,
            end=self.end_position,
            stroke=self.config.color,
            stroke_width=self.width,
        )
        return line



from dataclasses import dataclass
from itertools import product
from typing import List, Optional

from fretboardgtr.constants import CHORDS_DICT_ESSENTIAL, CHROMATICS_NOTES, SCALES_DICT
from fretboardgtr.utils import chromatic_position_from_root, get_note_from_index


def find_first_index(_list: List[int], value: int) -> Optional[int]:
    try:
        index = _list.index(value)
        return index
    except ValueError:
        return None  # If the value is not found in the tuple, return None


@dataclass
class NotesContainer:
    root: str
    notes: List[str]

    def get_probablely_possible_fingering(
        self, tuning: List[str]
    ) -> List[List[Optional[int]]]:
        """Get all probably possible fingering for a specific tuning.

        Parameters
        ----------
        tuning : List[str]
            List of note of the tuning

        Returns
        -------
        List[List[Optional[int]]]
            List of propably possible fingerings
        """
        scale = []
        for string_note in tuning:
            indices = []
            for note in self.notes:
                _idx = chromatic_position_from_root(note, string_note)
                while _idx <= 16:
                    indices.append(_idx)
                    _idx += 12
            scale.append(sorted(indices))

        fingerings = []
        for combination in product(*scale):
            non_zero_numbers = [num for num in combination if num != 0]

            # No more than 4 fingers but duplicated allowed
            if len(set(non_zero_numbers)) > 4:
                continue

            # No more than 5 frets spacing
            # else try to remplace min values by None
            # TODO: Also add the max checking
            new_combination = list(combination)

            while True:
                index_of_min = find_first_index(new_combination, min(non_zero_numbers))
                index_of_min_of_non_zero = find_first_index(
                    non_zero_numbers, min(non_zero_numbers)
                )

                if index_of_min is not None and index_of_min_of_non_zero is not None:
                    new_combination[index_of_min] = None
                    del non_zero_numbers[index_of_min_of_non_zero]

                if len(non_zero_numbers) < 2:
                    break

                if max(non_zero_numbers) - min(non_zero_numbers) <= 5:
                    break

            notes = []
            for index, note in zip(new_combination, tuning):
                if index is not None:
                    notes.append(get_note_from_index(index, note))

            # Each notes should appear at least once.
            if set(notes) != set(self.notes):
                continue
            fingerings.append(list(new_combination))
        return fingerings


class ScaleFromName:
    """Object that generating NotesContainer object from root and mode.

    Given a root name and a mode name, get the resulting scale.

    Also :
    Mode name can be given thanks to the constants.Mode enum as well as string
    Note name can be given thanks to the constants.Note enum as well as string

    Example
    -------
    >>> ScaleFromName(root='C',mode='Dorian').get()
        NotesContainer(root= 'C', scale = ['C', 'D', 'D#', 'F', 'G', 'A', 'A#'])
    >>> ScaleFromName(root=Note.C,mode=Mode.DORIAN).get()
        NotesContainer(root= 'C', scale = ['C', 'D', 'D#', 'F', 'G', 'A', 'A#'])
    """

    def __init__(self, root: str = "C", mode: str = "Ionian"):
        self.root = root
        self.mode = mode

    def get(self) -> NotesContainer:
        index = CHROMATICS_NOTES.index(self.root)
        mode_idx = SCALES_DICT[self.mode]
        scale = []
        for note_id in mode_idx:
            scale.append(CHROMATICS_NOTES[(index + note_id) % 12])
        return NotesContainer(self.root, scale)


class ChordFromName:
    """Object generating NotesContainer object from root and chord quality.

    Given a root name and a quality name, get the resulting scale.

    Also :
    Mode name can be given thanks to the constants.Chord enum as well as string
    Note name can be given thanks to the constants.Note enum as well as string

    Example
    -------
    >>> ChordFromName(root='C',quality='M').get()
        NotesContainer(root= 'C', scale = ['C', 'E', 'G'])
    >>> ChordFromName(root=Note.C,quality=Chord.MAJOR).resultget()
        NotesContainer(root= 'C', scale = ['C', 'E', 'G'])
    """

    def __init__(self, root: str = "C", quality: str = "M"):
        self.root = root
        self.quality = quality

    def get(self) -> NotesContainer:
        index = CHROMATICS_NOTES.index(self.root)

        quality_idx = CHORDS_DICT_ESSENTIAL[self.quality]
        scale = []
        for note_id in quality_idx:
            scale.append(CHROMATICS_NOTES[(index + note_id) % 12])
        return NotesContainer(self.root, scale)



from dataclasses import dataclass, field, fields
from typing import List, Optional, Union

from fretboardgtr.elements.background import Background
from fretboardgtr.elements.base import FretBoardElement
from fretboardgtr.elements.cross import Cross
from fretboardgtr.elements.fret_number import FretNumber
from fretboardgtr.elements.frets import Fret
from fretboardgtr.elements.neck_dots import NeckDot
from fretboardgtr.elements.notes import FrettedNote, OpenNote
from fretboardgtr.elements.nut import Nut
from fretboardgtr.elements.strings import String
from fretboardgtr.elements.tuning import Tuning


@dataclass
class FretBoardElements:
    """Container dataclass for the different elements of fretboards."""

    background: Optional[Background] = None
    fret_numbers: List[FretNumber] = field(default_factory=list)
    neck_dots: List[NeckDot] = field(default_factory=list)
    frets: List[Fret] = field(default_factory=list)
    nut: Optional[Nut] = None
    tuning: List[Tuning] = field(default_factory=list)
    strings: List[String] = field(default_factory=list)
    notes: List[Union[OpenNote, FrettedNote]] = field(default_factory=list)
    crosses: List[Cross] = field(default_factory=list)
    customs: List[FretBoardElement] = field(default_factory=list)

    def to_list(self) -> List[FretBoardElement]:
        """Convert the elements to a flat list of element."""
        flat_elements = []
        for element in fields(self):
            value = getattr(self, element.name)
            if isinstance(value, list):
                flat_elements.extend(value)
            elif value is not None:
                flat_elements.append(value)
        return flat_elements

    def __len__(self) -> int:
        return len(self.to_list())



from dataclasses import dataclass
from typing import Optional, Tuple

import svgwrite

from fretboardgtr.base import ConfigIniter
from fretboardgtr.constants import NO_COLOR
from fretboardgtr.elements.base import FretBoardElement


@dataclass
class BackgroundConfig(ConfigIniter):
    """Background element configuration."""

    color: str = NO_COLOR
    opacity: float = 0.7


class Background(FretBoardElement):
    """Background element to be drawn in the final fretboard."""

    def __init__(
        self,
        position: Tuple[float, float],
        size: Tuple[float, float],
        config: Optional[BackgroundConfig] = None,
    ):
        self.config = config if config else BackgroundConfig()
        self.position = position
        self.size = size

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the Background to a svgwrite object.

        This maps the BackgroundConfig configuration attributes to the
        svg attributes
        """
        rectangle = svgwrite.shapes.Rect(
            insert=self.position,
            size=self.size,  # -2 evite case du bas du tuning
            rx=None,
            ry=None,
            fill=self.config.color,
            fill_opacity=self.config.opacity,
        )
        return rectangle



from fretboardgtr.elements.frets import Fret, FretConfig


def test_fret_get_svg():
    fret = Fret(start_position=(0.0, 0.0), end_position=(10.0, 10.0))
    attribs = fret.get_svg().attribs
    print(attribs)
    assert attribs["x1"] == 0.0
    assert attribs["y1"] == 0.0
    assert attribs["x2"] == 10.0
    assert attribs["y2"] == 10.0


def test_fret_get_svg_custom_config():
    fret_config = FretConfig(color="blue")
    fret = Fret(
        start_position=(0.0, 0.0), end_position=(10.0, 10.0), config=fret_config
    )
    attribs = fret.get_svg().attribs
    assert attribs["x1"] == 0.0
    assert attribs["y1"] == 0.0
    assert attribs["x2"] == 10.0
    assert attribs["y2"] == 10.0
    assert attribs["stroke"] == "blue"



import copy
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

from fretboardgtr.constants import STANDARD_TUNING
from fretboardgtr.elements.background import Background
from fretboardgtr.elements.base import FretBoardElement
from fretboardgtr.elements.cross import Cross
from fretboardgtr.elements.fret_number import FretNumber
from fretboardgtr.elements.frets import Fret
from fretboardgtr.elements.neck_dots import NeckDot
from fretboardgtr.elements.notes import FrettedNote, OpenNote
from fretboardgtr.elements.nut import Nut
from fretboardgtr.elements.strings import String
from fretboardgtr.elements.tuning import Tuning
from fretboardgtr.exporters import EXPORTERS
from fretboardgtr.fretboards.base import FretBoardLike
from fretboardgtr.fretboards.config import FretBoardConfig
from fretboardgtr.fretboards.converters import FretBoardToSVGConverter
from fretboardgtr.fretboards.elements import FretBoardElements
from fretboardgtr.fretboards.horizontal import HorizontalFretBoard
from fretboardgtr.fretboards.vertical import VerticalFretBoard
from fretboardgtr.notes_creators import NotesContainer
from fretboardgtr.utils import (
    chromatic_position_from_root,
    get_note_from_index,
    get_valid_dots,
    note_to_interval_name,
)


def build_config(config: Optional[Union[Dict, FretBoardConfig]]) -> FretBoardConfig:
    if config is None:
        return FretBoardConfig()
    elif isinstance(config, dict):
        return FretBoardConfig.from_dict(config)
    elif isinstance(config, FretBoardConfig):
        return config
    else:
        msg = "Invalid config type {type(config)}."
        msg += "Need a dictionnary or FretBoardConfig instance"
        raise ValueError(msg)


class FretBoard(FretBoardLike):
    def __init__(
        self,
        tuning: Optional[List[str]] = None,
        config: Optional[Union[Dict, FretBoardConfig]] = None,
        vertical: bool = False,
    ):
        self.tuning = tuning if tuning is not None else STANDARD_TUNING

        self.config = build_config(config)
        self.elements = FretBoardElements()
        self.fretboard: Union[HorizontalFretBoard, VerticalFretBoard] = (
            HorizontalFretBoard(tuning, self.config)
            if not vertical
            else VerticalFretBoard(tuning, self.config)
        )
        self.init()

    def set_config(self, config: FretBoardConfig) -> None:
        self.config = config

    def init(self) -> None:
        """Init the fretboard by adding essential elements.

        This adds :
            background
            fret_numbers
            neck_dots
            frets
            nut
            tuning
            strings
        """
        self.add_background()
        self.add_fret_numbers()
        self.add_neck_dots()
        self.add_frets()
        self.add_nut()
        self.add_tuning()
        self.add_strings()

    def add_background(self) -> None:
        """Build and add background element."""
        x, y = self.fretboard.get_background_start_position()
        width, height = self.fretboard.get_background_dimensions()
        background = Background((x, y), (width, height), config=self.config.background)
        self.elements.background = background

    def add_neck_dots(self) -> None:
        """Build and add neck dot elements."""
        dots = get_valid_dots(
            self.config.general.first_fret, self.config.general.last_fret
        )
        for dot in dots:
            for position in self.fretboard.get_neck_dot_position(dot):
                center_dot = NeckDot(position, config=self.config.neck_dots)
                self.elements.neck_dots.append(center_dot)

    def add_frets(self) -> None:
        """Build and add fret elements."""
        show_nut = self.config.general.show_nut
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret

        # Skip the open virtual fret
        first_fret = 0
        if show_nut:
            first_fret = 1

        # +2 is to close the last fret of fretboard
        for fret_no in range(first_fret, number_of_frets + 2):
            position = self.fretboard.get_fret_position(fret_no)

            fret = Fret(position[0], position[1], config=self.config.frets)
            self.elements.frets.append(fret)

    def add_strings(self) -> None:
        """Build and add string elements."""
        # begin before if min fret !=0 because no open chords

        for string_no, _ in enumerate(self.tuning):
            start, end = self.fretboard.get_strings_position(string_no)
            string = String(start, end, config=self.config.strings)
            self.elements.strings.append(string)

    def add_nut(self) -> None:
        """Build and add nut element."""
        position = self.fretboard.get_nut_position()
        if not position:
            return None
        start, end = position
        nut = Nut(start, end, config=self.config.nut)
        self.elements.nut = nut

    def add_fret_numbers(self) -> None:
        """Build and add fret number elements."""
        if not self.config.general.show_frets:
            return None

        dots = get_valid_dots(
            self.config.general.first_fret, self.config.general.last_fret
        )
        for dot in dots:
            x, y = self.fretboard.get_fret_number_position(dot)
            fret_number = FretNumber(str(dot), (x, y), config=self.config.fret_numbers)
            self.elements.fret_numbers.append(fret_number)

    def add_tuning(self) -> None:
        """Build and add tuning element."""
        if not self.config.general.show_tuning:
            return None
        tuning = self.fretboard.get_list_in_good_order(self.tuning)
        for string_no, note in enumerate(tuning):
            x, y = self.fretboard.get_tuning_position(string_no)
            tuning_note = Tuning(note, (x, y), config=self.config.tuning)
            self.elements.tuning.append(tuning_note)

    def _get_open_note(
        self, position: Tuple[float, float], note: str, root: Optional[str] = None
    ) -> OpenNote:
        config = copy.copy(self.config.open_notes)
        if root and self.config.general.open_color_scale:
            idx = chromatic_position_from_root(note, root)
            color = self.config.general.open_colors.from_interval(idx)
            config.color = color

        if root and self.config.general.show_degree_name:
            note = note_to_interval_name(note, root)

        if not (
            self.config.general.show_degree_name or self.config.general.show_note_name
        ):
            note = ""

        _note = OpenNote(note, position, config=config)
        return _note

    def _get_fretted_note(
        self, position: Tuple[float, float], note: str, root: Optional[str] = None
    ) -> FrettedNote:
        config = copy.copy(self.config.fretted_notes)
        if root and self.config.general.fretted_color_scale:
            idx = chromatic_position_from_root(note, root)
            color = self.config.general.fretted_colors.from_interval(idx)
            config.color = color

        if root and self.config.general.show_degree_name:
            note = note_to_interval_name(note, root)

        if not (
            self.config.general.show_degree_name or self.config.general.show_note_name
        ):
            note = ""

        _note = FrettedNote(note, position, config=config)
        return _note

    def _add_single_note(
        self, string_no: int, index: int, note: str, root: Optional[str] = None
    ) -> None:
        position = self.fretboard.get_single_note_position(string_no, index)
        _note: Union[OpenNote, FrettedNote]
        if index == 0:
            _note = self._get_open_note(position, note, root)
        else:
            _note = self._get_fretted_note(position, note, root)
        self.elements.notes.append(_note)

    def add_note(self, string_no: int, note: str, root: Optional[str] = None) -> None:
        """Build and add background element."""
        if string_no < 0 or string_no > len(self.tuning):
            raise ValueError(f"String number is invalid. Tuning is {self.tuning}")

        # Note when fret == 0
        string_root = self.fretboard.get_list_in_good_order(self.tuning)[string_no]
        # Note when fret == N
        string_root = get_note_from_index(self.config.general.first_fret, string_root)
        # Index of the note from root
        _idx = chromatic_position_from_root(note, string_root)
        indices = []
        while _idx <= self.config.general.last_fret - self.config.general.first_fret:
            indices.append(_idx)
            _idx += 12

        for idx in indices:
            self._add_single_note(string_no, idx, note, root)

    def add_single_note_from_index(
        self, string_no: int, index: int, root: Optional[str] = None
    ) -> None:
        """Build and add background element."""
        string_note = self.tuning[string_no]
        note = get_note_from_index(index, string_note)

        self._add_single_note(string_no, index, note, root)

    def _add_cross_or_note(
        self, root: Optional[str], string_no: int, finger_position: Optional[int]
    ) -> None:
        if finger_position is None:
            position = self.fretboard.get_cross_position(string_no)
            cross = Cross(position, config=self.config.cross)
            self.elements.crosses.append(cross)
        else:
            string_note = get_note_from_index(
                finger_position,
                self.fretboard.get_list_in_good_order(self.tuning)[string_no],
            )
            if finger_position > 0:
                finger_position -= self.config.general.first_fret
                if finger_position <= 0:
                    return None

            self._add_single_note(string_no, finger_position, string_note, root)

    def add_notes(self, scale: NotesContainer) -> None:
        """Add an entire scale (from NoteContainer object) to the fretboard.

        Parameters
        ----------
        scale : NotesContainer
            Object representing the root and the associated scale
        """
        for string_no, _ in enumerate(self.tuning):
            for note in scale.notes:
                self.add_note(string_no, note, scale.root)

    def add_fingering(
        self, fingering: List[Optional[int]], root: Optional[str] = None
    ) -> None:
        """Add fingering starting with upper string to lower string.

        This function automatically calculate the note names or
        intervals if the root is given Display them in the way described
        in configuration
        """
        if len(fingering) != len(self.tuning):
            raise ValueError(
                f"Fingering has not the same size as tuning."
                f" Got {len(fingering)} expected {len(self.tuning)}"
            )
        for string_no, finger_position in enumerate(
            self.fretboard.get_list_in_good_order(fingering)
        ):
            self._add_cross_or_note(root, string_no, finger_position)

    def add_scale(
        self, scale: List[List[Optional[int]]], root: Optional[str] = None
    ) -> None:
        """Add scale starting with upper string to lower string.

        This function automatically calculate the note names or
        intervals if the root is given Display them in the way described
        in configuration
        """
        if len(scale) != len(self.tuning):
            raise ValueError(
                f"Scale has not the same size as tuning."
                f" Got {len(scale)} expected {len(self.tuning)}"
            )

        for string_no, finger_positions in enumerate(
            self.fretboard.get_list_in_good_order(scale)
        ):
            for finger_position in finger_positions:
                self._add_cross_or_note(root, string_no, finger_position)

    def add_note_element(self, note: Union[OpenNote, FrettedNote]) -> None:
        """Add a note element to the fretboard.

        Need to be either a OpenNote or a FrettedNote

        Parameters
        ----------
        note : Union[OpenNote, FrettedNote]
            Note element

        Raises
        ------
        ValueError
            If the note is not a Union[OpenNote, FrettedNote]
        """
        self.elements.notes.append(note)

    def add_element(self, element: FretBoardElement) -> None:
        """Add an element to the fretboard.

        Need to be either a FretBoardElement

        Parameters
        ----------
        element : FretBoardElement
            Fretboard element

        Raises
        ------
        ValueError
            If the element is not a FretboardElement
        """
        if not issubclass(type(element), FretBoardElement):
            raise ValueError("Element should be a 'FretBoardElement'")
        self.elements.customs.append(element)

    def get_elements(self) -> FretBoardElements:
        return self.elements

    def get_size(self) -> Tuple[float, float]:
        """Get total size of the drawing.

        Returns
        -------
        Tuple[float, float]
            Width and heigth
        """
        return self.fretboard.get_size()

    def get_inside_bounds(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Get the size of the inner drawing.

        This function could be use to add custom elements

        Returns
        -------
        Tuple[Tuple[float, float], Tuple[float, float]]
            Upper left corner x coordinate, upper left corner y coordinate
            Lower right corner x coordinate, lower right corner y coordinate
        """
        return self.fretboard.get_inside_bounds()

    def export(
        self, to: Union[Path, str], format: Literal["svg", "png", "pdf"] = "svg"
    ) -> None:
        """Export the fretboard to the desired path.

        Export availables are :
            svg
            png
            pdf

        Custom exporters can be created by :
        Subclassing : func:`~fretboardgtr.exporters.Exporter`
        Regitering it with func:`~fretboardgtr.exporters.register_exporter`

        Parameters
        ----------
        to : Union[Path, str]
            Path to export to
        format : Literal[&quot;svg&quot;, &quot;png&quot;, &quot;pdf&quot;], optional
            Format to export in, by default "svg"

        Raises
        ------
        NotImplementedError
            If the desired format is not implemented
        """
        to = Path(to)
        if format.upper() not in EXPORTERS:
            availables = ", ".join(list(EXPORTERS.keys()))
            raise NotImplementedError(
                f"Save to the {format} format is unsupported."
                f" Available formats are {availables}"
            )

        drawing = FretBoardToSVGConverter(self).convert()
        to.parent.mkdir(exist_ok=True)
        EXPORTERS[format.upper()](drawing).export(to=to)
